
# Encoder
import torch
import torch.nn as nn
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.preprocessing import MinMaxScaler

# Device configuration
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
latent_dim = 8
hidden_dim = 900
batch_size = 10000
learning_rate = 0.0008
num_epochs = 8000
# Encoder
class Encoder(nn.Module):
    def __init__(self, input_dim):
        super(Encoder, self).__init__()
        self.layers = nn.Sequential(
            nn.Linear(input_dim, 16),
            nn.ReLU(),
            nn.Linear(16, 32),
            nn.ReLU(),
            nn.Linear(32, 64),
            nn.ReLU(),
            nn.Linear(64, 32),
            nn.ReLU(),
            nn.Linear(32, 16),
            nn.ReLU(),
            nn.Linear(16, latent_dim)
        )

    def forward(self, x):
        return self.layers(x)

# Decoder
class Decoder(nn.Module):
    def __init__(self, output_dim):
        super(Decoder, self).__init__()
        self.layers = nn.Sequential(
            nn.Linear(latent_dim, 16),
            nn.ReLU(),
            nn.Linear(16, 32),
            nn.ReLU(),
            nn.Linear(32, 64),
            nn.ReLU(),
            nn.Linear(64, 32),
            nn.ReLU(),
            nn.Linear(32, 16),
            nn.ReLU(),
            nn.Linear(16, output_dim),
            nn.Tanh()
        )

    def forward(self, z):
        return self.layers(z)

# Load the trained model
def load_model(encoder, decoder, filename='trained_model.h5'):
    checkpoint = torch.load(filename)
    encoder.load_state_dict(checkpoint['encoder_state_dict'])
    decoder.load_state_dict(checkpoint['decoder_state_dict'])
    print(f'Model loaded from {filename}')

# Check synchronization function
# Check synchronization function
def check_synchronization(data, max_shift_amount, encoder, decoder):
    data = torch.tensor(data, dtype=torch.float32).to(device)
    best_shift_amount = None
    best_shifted_data = None
    best_reconstructed_data = None
    min_large_errors_count = float('inf')

    with torch.no_grad():
        z = encoder(data)
        recon = decoder(z)
        errors = torch.abs(recon - data)

    large_errors_count = (errors > 0.32).sum()
    if large_errors_count == 0:
        print("Data is synchronized. And no bad data exists.")
        return 0, data, recon

    for shift_amount in range(-max_shift_amount, max_shift_amount + 1):
        if shift_amount == 0:
            shifted_data = data
            compared_reconstructed_data = recon
        elif shift_amount < 0:
            shifted_data = data[:, :shift_amount]
            compared_reconstructed_data = recon[:, -shift_amount:]
        else:
            shifted_data = data[:, shift_amount:]
            compared_reconstructed_data = recon[:, :-shift_amount]

        errors = torch.abs(compared_reconstructed_data - shifted_data)
        large_errors_count = (errors > 0.32).sum()
        if large_errors_count / len(errors.T) <= 1 / 100:
            best_shift_amount = shift_amount
            best_shifted_data = shifted_data
            best_reconstructed_data = compared_reconstructed_data
        elif large_errors_count < min_large_errors_count:
            min_large_errors_count = large_errors_count

    if best_shift_amount is not None:
        print(f"Data is not synchronized. Shifting the data by {best_shift_amount} positions results in data back to synchronous.")
        best_errors = torch.abs(best_reconstructed_data - best_shifted_data)
        topk = 10
        max_errors, _ = torch.topk(best_errors, topk)
        min_errors, _ = torch.topk(-best_errors, topk)
        print(f'Max 10 errors for best shift amount: {max_errors}')
        print(f'Min 10 errors for best shift amount: {-min_errors}')
    else:
        print("Data is synchronized. Bad data detected.")

    return best_shift_amount, best_shifted_data, best_reconstructed_data

# Define the testing and post-processing function
def test_and_process(data_file='H:/SDE data/x_test_modified_new12allbad3PMUs.csv'):
    # Load and normalize the data
    data = pd.read_csv(data_file)
    time_data = np.arange(len(data))
    scaler = MinMaxScaler()
    data_normalized = pd.DataFrame(scaler.fit_transform(data), columns=data.columns)
    data_tensor = torch.tensor(data_normalized.values, dtype=torch.float32)

    # Store the original min and max values for each column for reverse normalization
    original_min_max_values = {column: (data[column].min(), data[column].max()) for column in data.columns}

    # Test data (test_indices will be defined as per your requirements)
    test_indices = [0, 2, 4, 6, 8, 10]
    test_data = data_tensor[:, test_indices]
    test_data_reshaped = test_data.transpose(0, 1)

    # Initialize models
    encoder = Encoder(test_data_reshaped.shape[1]).to(device)
    decoder = Decoder(test_data_reshaped.shape[1]).to(device)

    # Load trained model
    load_model(encoder, decoder, 'trained_model.h5')

    # Testing loop
    encoder.eval()
    decoder.eval()

    # Split the data into 6 sets
    test_data_sets = torch.chunk(test_data_reshaped, 6, dim=0)
    recon_sets = []

    for i, test_data_set in enumerate(test_data_sets):
        test_data_set = test_data_set.to(device)
        with torch.no_grad():
            z = encoder(test_data_set)
            recon = decoder(z)
            recon_sets.append(recon)

            # Compute overall loss
            loss = nn.MSELoss()(recon, test_data_set)
            print(f'Test Loss for PMU {i + 1}: {loss.item():.4f}')

            # Check synchronization for each set
            original_test_data_set = test_data_set.cpu().numpy()
            best_shift_amount, best_shifted_data, best_reconstructed_data = check_synchronization(
                original_test_data_set, 100, encoder, decoder)

            if best_shifted_data is not None:
                best_shifted_data_np = best_shifted_data.cpu().numpy()
                best_reconstructed_data_np = best_reconstructed_data.cpu().numpy()

                # Plot best shifted and reconstructed data
                fig, ax = plt.subplots(figsize=(10, 8))
                ax.plot(time_data[:best_shifted_data_np.shape[1]], best_shifted_data_np[0], label='Best Shifted Data')
                ax.plot(time_data[:best_reconstructed_data_np.shape[1]], best_reconstructed_data_np[0], linestyle='dashed',
                        label='Best Reconstructed Data')
                ax.set_xlabel('Time Index')
                ax.set_ylabel('Voltage Magnitude (p.u.)')
                ax.legend()
                plt.show()
            else:
                print("No best shifted data found.")

    # Initialize a dictionary to store bad_data_indices for each PMU
    bad_data_indices_dict = {}

    # Compute error for each sample in each set after the testing loop
    for i, (recon_set, test_data_set) in enumerate(zip(recon_sets, test_data_sets)):
        test_data_set = test_data_set.to(device)
        errors_set = torch.abs(recon_set - test_data_set)

        # Compute total error for each set
        total_error_set = torch.sum(errors_set)
        print(f'Total error for PMU {i + 1}: {total_error_set.item()}')

        # Print the 10 max and min errors for each set
        topk = 10
        max_errors_set, _ = torch.topk(errors_set, topk)
        min_errors_set, _ = torch.topk(-errors_set, topk)
        print(f'Max 10 errors for PMU {i + 1}: {max_errors_set}')
        print(f'Min 10 errors for PMU {i + 1}: {-min_errors_set}')

        # Detect bad data
        large_errors_count = (errors_set > 0.29).sum()
        if large_errors_count == 0:
            print(f"No bad data detected in PMU {i + 1}")
        else:
            print(f"Bad data detected in PMU {i + 1}")
            bad_data_indices = (errors_set > 0.29).nonzero(as_tuple=True)[1]
            print(f"Bad data located at data sample: {bad_data_indices.tolist()}")
            bad_data_indices_dict[i] = bad_data_indices.tolist()

    # Concatenate all recon tensors along dimension 1
    recon_all = torch.cat(recon_sets, dim=0)

    # Reverse normalization for original and reconstructed test data
    original_test_data = test_data_reshaped.cpu().numpy()
    reconstructed_test_data = recon_all.cpu().numpy()

    original_test_data_df = pd.DataFrame(original_test_data.transpose(), columns=data.columns[test_indices])
    reconstructed_test_data_df = pd.DataFrame(reconstructed_test_data.transpose(), columns=data.columns[test_indices])

    for column in original_test_data_df.columns:
        min_val, max_val = original_min_max_values[column]
        original_test_data_df[column] = original_test_data_df[column] * (max_val - min_val) + min_val

    for column in reconstructed_test_data_df.columns:
        min_val, max_val = original_min_max_values[column]
        reconstructed_test_data_df[column] = reconstructed_test_data_df[column] * (max_val - min_val) + min_val

    reversed_original_test_data = original_test_data_df.values.transpose()
    reversed_reconstructed_test_data = reconstructed_test_data_df.values.transpose()

    # Plot testing data after reverse normalization
    fig_test_after, ax_test_after = plt.subplots(figsize=(10, 8))

    for i in range(3):
        ax_test_after.plot(time_data, reversed_original_test_data[i], label=f'Original PMU {test_indices[i] + 1} data')
        ax_test_after.plot(time_data, reversed_reconstructed_test_data[i], linestyle='dashed', label=f'Reconstructed PMU {test_indices[i] + 1} data')

    ax_test_after.set_xlabel('Time Index')
    ax_test_after.set_ylabel('Voltage Magnitude (p.u.)')
    ax_test_after.legend(ncol=2)
    plt.show()

    # Iterate over the bad_data_indices_dict to replace bad data
    for pmu, bad_data_indices in bad_data_indices_dict.items():
        column = data.columns[(pmu) * 2]
        data.loc[bad_data_indices, column] = reconstructed_test_data_df.loc[bad_data_indices, column]

    # Save the modified DataFrame to a new CSV file
    output_file_path = 'H:/SDE data/x_test_reconstructed_new12allbad3PMUs_a.csv'
    data.to_csv(output_file_path, index=False)
    print("Data processing complete. Output saved to:", output_file_path)

# Call the test function
test_and_process()